@@ -7,6 +7,8 @@
 #include "tool_rc.h"
 #include "tpm2.h"
 #include "tpm2_auth_util.h"
+#include "tpm2_tool.h"
+#include "config.h"
 
 #define TPM2_ERROR_TSS2_RC_ERROR_MASK 0xFFFF
 
@@ -15,12 +17,12 @@ static inline UINT16 tpm2_error_get(TSS2_RC rc) {
 }
 
 tool_rc tpm2_readpublic(ESYS_CONTEXT *esys_context, ESYS_TR object_handle,
-        ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3,
         TPM2B_PUBLIC **out_public, TPM2B_NAME **name,
         TPM2B_NAME **qualified_name) {
 
-    TSS2_RC rval = Esys_ReadPublic(esys_context, object_handle, shandle1,
-            shandle2, shandle3, out_public, name, qualified_name);
+    TSS2_RC rval = Esys_ReadPublic(esys_context, object_handle,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
+            out_public, name, qualified_name);
     if (rval != TPM2_RC_SUCCESS) {
         LOG_PERR(Esys_ReadPublic, rval);
         return tool_rc_from_tpm(rval);
@@ -92,11 +94,10 @@ tool_rc tpm2_close(ESYS_CONTEXT *esys_context, ESYS_TR *rsrc_handle) {
 }
 
 tool_rc tpm2_nv_readpublic(ESYS_CONTEXT *esys_context, ESYS_TR nv_index,
-        ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3,
         TPM2B_NV_PUBLIC **nv_public, TPM2B_NAME **nv_name) {
 
-    TSS2_RC rval = Esys_NV_ReadPublic(esys_context, nv_index, shandle1, shandle2,
-            shandle3, nv_public, nv_name);
+    TSS2_RC rval = Esys_NV_ReadPublic(esys_context, nv_index,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, nv_public, nv_name);
 
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_NV_ReadPublic, rval);
@@ -106,12 +107,11 @@ tool_rc tpm2_nv_readpublic(ESYS_CONTEXT *esys_context, ESYS_TR nv_index,
     return tool_rc_success;
 }
 
-tool_rc tpm2_getcap(ESYS_CONTEXT *esys_context, ESYS_TR shandle1,
-        ESYS_TR shandle2, ESYS_TR shandle3, TPM2_CAP capability,
+tool_rc tpm2_getcap(ESYS_CONTEXT *esys_context, TPM2_CAP capability,
         UINT32 property, UINT32 property_count, TPMI_YES_NO *more_data,
         TPMS_CAPABILITY_DATA **capability_data) {
 
-    TSS2_RC rval = Esys_GetCapability(esys_context, shandle1, shandle2, shandle3,
+    TSS2_RC rval = Esys_GetCapability(esys_context, ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
             capability, property, property_count, more_data, capability_data);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_NV_ReadPublic, rval);
@@ -211,6 +211,19 @@ tool_rc tpm2_sess_get_attributes(ESYS_CONTEXT *esys_context, ESYS_TR session,
     return tool_rc_success;
 }
 
+tool_rc tpm2_sess_get_noncetpm(ESYS_CONTEXT *esys_context,
+    ESYS_TR session_handle, TPM2B_NONCE **nonce_tpm) {
+
+    TSS2_RC rval = Esys_TRSess_GetNonceTPM(esys_context, session_handle,
+        nonce_tpm);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_TRSess_GetNonceTPM, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_policy_restart(ESYS_CONTEXT *esys_context, ESYS_TR session_handle,
         ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3) {
 
@@ -304,6 +317,32 @@ tool_rc tpm2_policy_or(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
     return tool_rc_success;
 }
 
+tool_rc tpm2_policy_namehash(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
+    const TPM2B_DIGEST *name_hash) {
+
+    TSS2_RC rval = Esys_PolicyNameHash(esys_context, policy_session,
+        ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, name_hash);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyNameHash, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_template(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
+    const TPM2B_DIGEST *template_hash) {
+
+    TSS2_RC rval = Esys_PolicyTemplate(esys_context, policy_session,
+        ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, template_hash);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyTemplate, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_policy_pcr(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
         ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3,
         const TPM2B_DIGEST *pcr_digest, const TPML_PCR_SELECTION *pcrs) {
@@ -331,15 +370,137 @@ tool_rc tpm2_policy_password(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
     return tool_rc_success;
 }
 
+tool_rc tpm2_policy_signed(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_entity_obj, ESYS_TR policy_session,
+        const TPMT_SIGNATURE *signature, INT32 expiration,
+        TPM2B_TIMEOUT **timeout, TPMT_TK_AUTH **policy_ticket,
+        TPM2B_NONCE *policy_qualifier, TPM2B_NONCE *nonce_tpm) {
+
+    TSS2_RC rval = Esys_PolicySigned(esys_context, auth_entity_obj->tr_handle,
+        policy_session, ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, nonce_tpm,
+        NULL, policy_qualifier, expiration, signature, timeout, policy_ticket);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicySigned, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_ticket(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
+    const TPM2B_TIMEOUT *timeout, const TPM2B_NONCE *policyref,
+    const TPM2B_NAME *authname, const TPMT_TK_AUTH *ticket) {
+
+    TSS2_RC rval = Esys_PolicyTicket(esys_context, policy_session, ESYS_TR_NONE,
+        ESYS_TR_NONE, ESYS_TR_NONE, timeout, NULL, policyref, authname, ticket);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicySigned, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_authvalue(ESYS_CONTEXT *esys_context,
+        ESYS_TR policy_session, ESYS_TR shandle1, ESYS_TR shandle2,
+        ESYS_TR shandle3) {
+
+    TSS2_RC rval = Esys_PolicyAuthValue(esys_context, policy_session, shandle1,
+            shandle2, shandle3);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyAuthValue, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_authorize_nv(ESYS_CONTEXT *esys_context,
+    tpm2_loaded_object *auth_entity_obj, TPM2_HANDLE nv_index,
+    ESYS_TR policy_session) {
+
+    ESYS_TR esys_tr_nv_index;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_index);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_entity_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_entity_obj->tr_handle, auth_entity_obj->session,
+            &auth_entity_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get auth entity obj session");
+        return rc;
+    }
+
+    rval = Esys_PolicyAuthorizeNV(esys_context, auth_entity_obj->tr_handle,
+        esys_tr_nv_index, policy_session, auth_entity_obj_session_handle,
+        ESYS_TR_NONE, ESYS_TR_NONE);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyAuthorizeNV, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_nv(ESYS_CONTEXT *esys_context,
+    tpm2_loaded_object *auth_entity_obj, TPM2_HANDLE nv_index,
+    ESYS_TR policy_session, const TPM2B_OPERAND *operand_b, UINT16 offset,
+    TPM2_EO operation) {
+
+    ESYS_TR esys_tr_nv_index;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_index);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_entity_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_entity_obj->tr_handle, auth_entity_obj->session,
+            &auth_entity_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get auth entity obj session");
+        return rc;
+    }
+
+    rval = Esys_PolicyNV(esys_context, auth_entity_obj->tr_handle,
+        esys_tr_nv_index, policy_session, auth_entity_obj_session_handle,
+        ESYS_TR_NONE, ESYS_TR_NONE, operand_b, offset, operation);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyNV, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_countertimer(ESYS_CONTEXT *esys_context,
+    ESYS_TR policy_session, const TPM2B_OPERAND *operand_b, UINT16 offset,
+    TPM2_EO operation) {
+
+    TSS2_RC rval = Esys_PolicyCounterTimer(esys_context, policy_session,
+        ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, operand_b, offset, operation);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyCounterTimer, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_policy_secret(ESYS_CONTEXT *esys_context,
-        tpm2_loaded_object *auth_entity_obj, ESYS_TR policy_session) {
+        tpm2_loaded_object *auth_entity_obj, ESYS_TR policy_session,
+        INT32 expiration, TPMT_TK_AUTH **policy_ticket,
+        TPM2B_TIMEOUT **timeout, TPM2B_NONCE *nonce_tpm,
+        TPM2B_NONCE *policy_qualifier) {
 
-    const TPM2B_NONCE *nonce_tpm = NULL;
     const TPM2B_DIGEST *cp_hash_a = NULL;
-    const TPM2B_NONCE *policy_ref = NULL;
-    INT32 expiration = 0;
-    TPM2B_TIMEOUT **timeout = NULL;
-    TPMT_TK_AUTH **policy_ticket = NULL;
 
     ESYS_TR auth_entity_obj_session_handle = ESYS_TR_NONE;
     tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
@@ -352,7 +513,7 @@ tool_rc tpm2_policy_secret(ESYS_CONTEXT *esys_context,
 
     TSS2_RC rval = Esys_PolicySecret(esys_context, auth_entity_obj->tr_handle,
             policy_session, auth_entity_obj_session_handle, ESYS_TR_NONE,
-            ESYS_TR_NONE, nonce_tpm, cp_hash_a, policy_ref, expiration, timeout,
+            ESYS_TR_NONE, nonce_tpm, cp_hash_a, policy_qualifier, expiration, timeout,
             policy_ticket);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_PolicySecret, rval);
@@ -390,6 +551,20 @@ tool_rc tpm2_policy_command_code(ESYS_CONTEXT *esys_context,
     return tool_rc_success;
 }
 
+tool_rc tpm2_policy_nv_written(ESYS_CONTEXT *esys_context,
+        ESYS_TR policy_session, ESYS_TR shandle1, ESYS_TR shandle2,
+        ESYS_TR shandle3, TPMI_YES_NO written_set) {
+
+    TSS2_RC rval = Esys_PolicyNvWritten(esys_context, policy_session, shandle1,
+            shandle2, shandle3, written_set);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PolicyNVWritten, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_policy_locality(ESYS_CONTEXT *esys_context, ESYS_TR policy_session,
         ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3,
         TPMA_LOCALITY locality) {
@@ -469,13 +644,44 @@ tool_rc tpm2_evictcontrol(ESYS_CONTEXT *esys_context,
     return tool_rc_success;
 }
 
+/* This function addresses ESAPI change that changes parameter type from
+ * Esys_TR to TPMI_RH_HIERARCHY and breaks backwards compatibility.
+ * To keep the tools parameters consistent after v4.0 release we need to
+ * map the values to appropriate type based on the version of the ESYS API.
+ * Note: the mapping is based on the ESYS version recognized at compile time.
+ * The TSS change can be found here:
+ * https://github.com/tpm2-software/tpm2-tss/pull/1531
+ */
+uint32_t fix_esys_hierarchy(TPMI_RH_HIERARCHY hierarchy)
+{
+#if defined(ESYS_3_0)
+    switch (hierarchy) {
+        case TPM2_RH_NULL:
+            return ESYS_TR_RH_NULL;
+        case TPM2_RH_OWNER:
+            return ESYS_TR_RH_OWNER;
+        case TPM2_RH_ENDORSEMENT:
+            return ESYS_TR_RH_ENDORSEMENT;
+        case TPM2_RH_PLATFORM:
+            return ESYS_TR_RH_PLATFORM;
+        default:
+            return TSS2_ESYS_RC_BAD_VALUE;
+    }
+#elif defined(ESYS_2_3)
+    return hierarchy;
+#else
+    UNUSED(hierarchy);
+    return TSS2_ESYS_RC_BAD_VALUE;
+#endif
+}
+
 tool_rc tpm2_hash(ESYS_CONTEXT *esys_context, ESYS_TR shandle1, ESYS_TR shandle2,
         ESYS_TR shandle3, const TPM2B_MAX_BUFFER *data, TPMI_ALG_HASH hash_alg,
         TPMI_RH_HIERARCHY hierarchy, TPM2B_DIGEST **out_hash,
         TPMT_TK_HASHCHECK **validation) {
 
     TSS2_RC rval = Esys_Hash(esys_context, shandle1, shandle2, shandle3, data,
-            hash_alg, hierarchy, out_hash, validation);
+            hash_alg, fix_esys_hierarchy(hierarchy), out_hash, validation);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_Hash, rval);
         return tool_rc_from_tpm(rval);
@@ -484,26 +690,24 @@ tool_rc tpm2_hash(ESYS_CONTEXT *esys_context, ESYS_TR shandle1, ESYS_TR shandle2
     return tool_rc_success;
 }
 
-tool_rc tpm2_hash_sequence_start(ESYS_CONTEXT *esys_context, ESYS_TR shandle1,
-        ESYS_TR shandle2, ESYS_TR shandle3, const TPM2B_AUTH *auth,
+tool_rc tpm2_hash_sequence_start(ESYS_CONTEXT *esys_context, const TPM2B_AUTH *auth,
         TPMI_ALG_HASH hash_alg, ESYS_TR *sequence_handle) {
 
-    TSS2_RC rval = Esys_HashSequenceStart(esys_context, shandle1, shandle2,
-            shandle3, auth, hash_alg, sequence_handle);
+    TSS2_RC rval = Esys_HashSequenceStart(esys_context, ESYS_TR_NONE, ESYS_TR_NONE,
+            ESYS_TR_NONE, auth, hash_alg, sequence_handle);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_HashSequenceStart, rval);
         return tool_rc_from_tpm(rval);
     }
 
-    return tool_rc_success;
+    return tpm2_tr_set_auth(esys_context, *sequence_handle, auth);
 }
 
 tool_rc tpm2_sequence_update(ESYS_CONTEXT *esys_context, ESYS_TR sequence_handle,
-        ESYS_TR shandle1, ESYS_TR shandle2, ESYS_TR shandle3,
         const TPM2B_MAX_BUFFER *buffer) {
 
-    TSS2_RC rval = Esys_SequenceUpdate(esys_context, sequence_handle, shandle1,
-            shandle2, shandle3, buffer);
+    TSS2_RC rval = Esys_SequenceUpdate(esys_context, sequence_handle, ESYS_TR_PASSWORD,
+            ESYS_TR_NONE, ESYS_TR_NONE, buffer);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_SequenceUpdate, rval);
         return tool_rc_from_tpm(rval);
@@ -513,13 +717,13 @@ tool_rc tpm2_sequence_update(ESYS_CONTEXT *esys_context, ESYS_TR sequence_handle
 }
 
 tool_rc tpm2_sequence_complete(ESYS_CONTEXT *esys_context,
-        ESYS_TR sequence_handle, ESYS_TR shandle1, ESYS_TR shandle2,
-        ESYS_TR shandle3, const TPM2B_MAX_BUFFER *buffer,
+        ESYS_TR sequence_handle, const TPM2B_MAX_BUFFER *buffer,
         TPMI_RH_HIERARCHY hierarchy, TPM2B_DIGEST **result,
         TPMT_TK_HASHCHECK **validation) {
 
-    TSS2_RC rval = Esys_SequenceComplete(esys_context, sequence_handle, shandle1,
-            shandle2, shandle3, buffer, hierarchy, result, validation);
+    TSS2_RC rval = Esys_SequenceComplete(esys_context, sequence_handle,
+            ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE, buffer,
+            fix_esys_hierarchy(hierarchy), result, validation);
     if (rval != TSS2_RC_SUCCESS) {
         LOG_PERR(Esys_SequenceComplete, rval);
         return tool_rc_from_tpm(rval);
@@ -528,6 +732,27 @@ tool_rc tpm2_sequence_complete(ESYS_CONTEXT *esys_context,
     return tool_rc_success;
 }
 
+tool_rc tpm2_event_sequence_complete(ESYS_CONTEXT *ectx, ESYS_TR pcr,
+        ESYS_TR sequence_handle, tpm2_session *session,
+        const TPM2B_MAX_BUFFER *buffer, TPML_DIGEST_VALUES **results) {
+
+    ESYS_TR shandle1 = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx, pcr, session,
+            &shandle1);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_EventSequenceComplete(ectx, pcr, sequence_handle, shandle1,
+            ESYS_TR_PASSWORD, ESYS_TR_NONE, buffer, results);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_EventSequenceComplete, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_tr_set_auth(ESYS_CONTEXT *esys_context, ESYS_TR handle,
         TPM2B_AUTH const *auth_value) {
 
@@ -741,6 +966,21 @@ tool_rc tpm2_rsa_decrypt(ESYS_CONTEXT *ectx, tpm2_loaded_object *keyobj,
     return tool_rc_success;
 }
 
+tool_rc tpm2_rsa_encrypt(ESYS_CONTEXT *ectx, tpm2_loaded_object *keyobj,
+        const TPM2B_PUBLIC_KEY_RSA *message, const TPMT_RSA_DECRYPT *scheme,
+        const TPM2B_DATA *label, TPM2B_PUBLIC_KEY_RSA **cipher_text) {
+
+    TSS2_RC rval = Esys_RSA_Encrypt(ectx, keyobj->tr_handle,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, message, scheme,
+            label, cipher_text);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_RSA_Encrypt, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_load(ESYS_CONTEXT *esys_context, tpm2_loaded_object *parentobj,
         const TPM2B_PRIVATE *in_private, const TPM2B_PUBLIC *in_public,
         ESYS_TR *object_handle) {
@@ -1126,7 +1366,6 @@ tool_rc tpm2_nvreadlock(ESYS_CONTEXT *esys_context,
             esys_tr_nv_handle, auth_hierarchy_obj_session_handle, ESYS_TR_NONE,
             ESYS_TR_NONE);
     if (rval != TPM2_RC_SUCCESS) {
-        LOG_ERR("Failed to lock NVRAM area at index 0x%X", nv_index);
         LOG_PERR(Esys_NV_ReadLock, rval);
         return tool_rc_from_tpm(rval);
     }
@@ -1134,6 +1373,123 @@ tool_rc tpm2_nvreadlock(ESYS_CONTEXT *esys_context,
     return tool_rc_success;
 }
 
+tool_rc tpm2_nvwritelock(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nv_index) {
+
+    ESYS_TR esys_tr_nv_handle;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_handle);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_hierarchy_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_hierarchy_obj->tr_handle, auth_hierarchy_obj->session,
+            &auth_hierarchy_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get shandle");
+        return rc;
+    }
+
+    rval = Esys_NV_WriteLock(esys_context, auth_hierarchy_obj->tr_handle,
+            esys_tr_nv_handle, auth_hierarchy_obj_session_handle, ESYS_TR_NONE,
+            ESYS_TR_NONE);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_NV_WriteLock, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_nvglobalwritelock(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_hierarchy_obj) {
+
+    ESYS_TR auth_hierarchy_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_hierarchy_obj->tr_handle, auth_hierarchy_obj->session,
+            &auth_hierarchy_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get shandle");
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_NV_GlobalWriteLock(esys_context, auth_hierarchy_obj->tr_handle,
+            auth_hierarchy_obj_session_handle, ESYS_TR_NONE, ESYS_TR_NONE);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_NV_GlobalWriteLock, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_nvsetbits(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nv_index,
+        UINT64 bits) {
+
+    ESYS_TR esys_tr_nv_handle;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_handle);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_hierarchy_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_hierarchy_obj->tr_handle, auth_hierarchy_obj->session,
+            &auth_hierarchy_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get shandle");
+        return rc;
+    }
+
+    rval = Esys_NV_SetBits(esys_context, auth_hierarchy_obj->tr_handle,
+            esys_tr_nv_handle, auth_hierarchy_obj_session_handle, ESYS_TR_NONE,
+            ESYS_TR_NONE, bits);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_NV_SetBits, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_nvextend(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nv_index,
+        TPM2B_MAX_NV_BUFFER *data) {
+
+    ESYS_TR esys_tr_nv_handle;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_handle);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_hierarchy_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_hierarchy_obj->tr_handle, auth_hierarchy_obj->session,
+            &auth_hierarchy_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Failed to get shandle");
+        return rc;
+    }
+
+    rval = Esys_NV_Extend(esys_context, auth_hierarchy_obj->tr_handle,
+            esys_tr_nv_handle, auth_hierarchy_obj_session_handle, ESYS_TR_NONE,
+            ESYS_TR_NONE, data);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_NV_Extend, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_nvundefine(ESYS_CONTEXT *esys_context,
         tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nv_index) {
 
@@ -1168,6 +1524,46 @@ tool_rc tpm2_nvundefine(ESYS_CONTEXT *esys_context,
     return tool_rc_success;
 }
 
+tool_rc tpm2_nvundefinespecial(ESYS_CONTEXT *esys_context,
+        tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nv_index,
+        tpm2_session *policy_session) {
+
+    ESYS_TR esys_tr_nv_handle;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_handle);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    ESYS_TR auth_hierarchy_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            auth_hierarchy_obj->tr_handle, auth_hierarchy_obj->session,
+            &auth_hierarchy_obj_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Couldn't get shandle");
+        return rc;
+    }
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(policy_session);
+
+    rval = Esys_NV_UndefineSpaceSpecial(esys_context,
+            esys_tr_nv_handle,
+            auth_hierarchy_obj->tr_handle,
+            policy_session_handle, // policy session
+            auth_hierarchy_obj_session_handle, // auth session for hierarchy
+            ESYS_TR_NONE);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_ERR("Failed to release NV area at index 0x%X", nv_index);
+        LOG_PERR(Esys_NV_UndefineSpace, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    LOG_INFO("Success to release NV area at index 0x%x.", nv_index);
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_nvwrite(ESYS_CONTEXT *esys_context,
         tpm2_loaded_object *auth_hierarchy_obj, TPM2_HANDLE nvindex,
         const TPM2B_MAX_NV_BUFFER *data, UINT16 offset) {
@@ -1266,6 +1662,98 @@ tool_rc tpm2_sign(ESYS_CONTEXT *esys_context, tpm2_loaded_object *signingkey_obj
     return tool_rc_success;
 }
 
+tool_rc tpm2_nvcertify(ESYS_CONTEXT *esys_context,
+    tpm2_loaded_object *signingkey_obj, tpm2_loaded_object *nvindex_authobj,
+    TPM2_HANDLE nv_index, UINT16 offset, UINT16 size,
+    TPMT_SIG_SCHEME *in_scheme, TPM2B_ATTEST **certify_info,
+    TPMT_SIGNATURE **signature, TPM2B_DATA *policy_qualifier) {
+
+    ESYS_TR signingkey_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            signingkey_obj->tr_handle, signingkey_obj->session,
+            &signingkey_obj_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    ESYS_TR nvindex_authobj_session_handle = ESYS_TR_NONE;
+    rc = tpm2_auth_util_get_shandle(esys_context,
+        nvindex_authobj->tr_handle, nvindex_authobj->session,
+        &nvindex_authobj_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    ESYS_TR esys_tr_nv_index;
+    TSS2_RC rval = Esys_TR_FromTPMPublic(esys_context, nv_index, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE, &esys_tr_nv_index);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_TR_FromTPMPublic, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    rval = Esys_NV_Certify(esys_context, signingkey_obj->tr_handle,
+        nvindex_authobj->tr_handle, esys_tr_nv_index,
+        signingkey_obj_session_handle, nvindex_authobj_session_handle,
+        ESYS_TR_NONE, policy_qualifier, in_scheme, size, offset, certify_info,
+        signature);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_NV_Certify, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_certifycreation(ESYS_CONTEXT *esys_context,
+    tpm2_loaded_object *signingkey_obj, tpm2_loaded_object *certifiedkey_obj,
+    TPM2B_DIGEST *creation_hash, TPMT_SIG_SCHEME *in_scheme,
+    TPMT_TK_CREATION *creation_ticket, TPM2B_ATTEST **certify_info,
+    TPMT_SIGNATURE **signature, TPM2B_DATA *policy_qualifier) {
+
+    ESYS_TR signingkey_obj_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            signingkey_obj->tr_handle, signingkey_obj->session,
+            &signingkey_obj_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_CertifyCreation(esys_context, signingkey_obj->tr_handle,
+        certifiedkey_obj->tr_handle, signingkey_obj_session_handle,
+        ESYS_TR_NONE, ESYS_TR_NONE, policy_qualifier, creation_hash, in_scheme,
+        creation_ticket, certify_info, signature);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_CertifyCreation, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_setprimarypolicy(ESYS_CONTEXT *esys_context,
+    tpm2_loaded_object *hierarchy_object, TPM2B_DIGEST *auth_policy,
+    TPMI_ALG_HASH hash_algorithm) {
+
+    ESYS_TR hierarchy_object_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(esys_context,
+            hierarchy_object->tr_handle, hierarchy_object->session,
+            &hierarchy_object_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_SetPrimaryPolicy(esys_context,
+        hierarchy_object->tr_handle, hierarchy_object_session_handle,
+        ESYS_TR_NONE, ESYS_TR_NONE, auth_policy, hash_algorithm);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_SetPrimaryPolicy, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_quote(ESYS_CONTEXT *esys_context, tpm2_loaded_object *quote_obj,
         TPMT_SIG_SCHEME *in_scheme, TPM2B_DATA *qualifying_data,
         TPML_PCR_SELECTION *pcr_select, TPM2B_ATTEST **quoted,
@@ -1290,6 +1778,46 @@ tool_rc tpm2_quote(ESYS_CONTEXT *esys_context, tpm2_loaded_object *quote_obj,
     return tool_rc_success;
 }
 
+tool_rc tpm2_changeeps(ESYS_CONTEXT *ectx,
+    tpm2_session *platform_hierarchy_session) {
+
+    ESYS_TR platform_hierarchy_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx, ESYS_TR_RH_PLATFORM,
+        platform_hierarchy_session, &platform_hierarchy_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_ChangeEPS(ectx, ESYS_TR_RH_PLATFORM,
+        platform_hierarchy_session_handle, ESYS_TR_NONE, ESYS_TR_NONE);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_ChangeEPS, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_changepps(ESYS_CONTEXT *ectx,
+    tpm2_session *platform_hierarchy_session) {
+
+    ESYS_TR platform_hierarchy_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx, ESYS_TR_RH_PLATFORM,
+        platform_hierarchy_session, &platform_hierarchy_session_handle);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_ChangePPS(ectx, ESYS_TR_RH_PLATFORM,
+        platform_hierarchy_session_handle, ESYS_TR_NONE, ESYS_TR_NONE);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_ChangePPS, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
 tool_rc tpm2_unseal(ESYS_CONTEXT *esys_context, tpm2_loaded_object *sealkey_obj,
         TPM2B_SENSITIVE_DATA **out_data) {
 
@@ -1309,3 +1837,272 @@ tool_rc tpm2_unseal(ESYS_CONTEXT *esys_context, tpm2_loaded_object *sealkey_obj,
 
     return tool_rc_success;
 }
+
+tool_rc tpm2_incrementalselftest(ESYS_CONTEXT *ectx, const TPML_ALG *to_test,
+        TPML_ALG **to_do_list) {
+
+    TSS2_RC rval = Esys_IncrementalSelfTest(ectx, ESYS_TR_NONE, ESYS_TR_NONE,
+        ESYS_TR_NONE, to_test, to_do_list);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_IncrementalSelfTest, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_stirrandom(ESYS_CONTEXT *ectx,
+        const TPM2B_SENSITIVE_DATA *data) {
+
+    TSS2_RC rval = Esys_StirRandom(ectx, ESYS_TR_NONE, ESYS_TR_NONE,
+        ESYS_TR_NONE, data);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_StirRandom, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_selftest(ESYS_CONTEXT *ectx, TPMI_YES_NO full_test) {
+
+    TSS2_RC rval = Esys_SelfTest(ectx, ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
+        full_test);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_SelfTest, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_gettestresult(ESYS_CONTEXT *ectx, TPM2B_MAX_BUFFER **out_data,
+        TPM2_RC *test_result) {
+
+    TSS2_RC rval = Esys_GetTestResult(ectx, ESYS_TR_NONE, ESYS_TR_NONE,
+            ESYS_TR_NONE, out_data, test_result);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_GetTestResult, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_loadexternal(ESYS_CONTEXT *ectx, const TPM2B_SENSITIVE *private,
+        const TPM2B_PUBLIC *public, TPMI_RH_HIERARCHY hierarchy,
+        ESYS_TR *object_handle) {
+
+    TSS2_RC rval = Esys_LoadExternal(ectx,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
+            private, public, hierarchy,
+            object_handle);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_LoadExternal, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_pcr_event(ESYS_CONTEXT *ectx,
+        ESYS_TR pcr, tpm2_session *session,
+        const TPM2B_EVENT *event_data,
+        TPML_DIGEST_VALUES **digests) {
+
+    ESYS_TR shandle1 = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx, pcr, session,
+            &shandle1);
+    if (rc != tool_rc_success) {
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_PCR_Event(ectx, pcr, shandle1, ESYS_TR_NONE,
+            ESYS_TR_NONE, event_data, digests);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PCR_Event, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_getrandom(ESYS_CONTEXT *ectx, UINT16 count,
+        TPM2B_DIGEST **random) {
+
+    TSS2_RC rval = Esys_GetRandom(ectx, ESYS_TR_NONE, ESYS_TR_NONE,
+        ESYS_TR_NONE, count, random);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_GetRandom, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_startup(ESYS_CONTEXT *ectx, TPM2_SU startup_type) {
+
+    TSS2_RC rval = Esys_Startup(ectx, startup_type);
+    if (rval != TPM2_RC_SUCCESS && rval != TPM2_RC_INITIALIZE) {
+        LOG_PERR(Esys_Startup, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_pcr_reset(ESYS_CONTEXT *ectx, ESYS_TR pcr_handle) {
+
+    TSS2_RC rval = Esys_PCR_Reset(ectx, pcr_handle, ESYS_TR_PASSWORD,
+            ESYS_TR_NONE, ESYS_TR_NONE);
+    if (rval != TSS2_RC_SUCCESS) {
+        LOG_PERR(Esys_PCR_Reset, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_makecredential(ESYS_CONTEXT *ectx, ESYS_TR handle,
+        const TPM2B_DIGEST *credential, const TPM2B_NAME *object_name,
+        TPM2B_ID_OBJECT **credential_blob, TPM2B_ENCRYPTED_SECRET **secret) {
+
+    TSS2_RC rval = Esys_MakeCredential(ectx, handle, ESYS_TR_NONE, ESYS_TR_NONE,
+            ESYS_TR_NONE, credential, object_name, credential_blob,
+            secret);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_MakeCredential, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_verifysignature(ESYS_CONTEXT *ectx, ESYS_TR key_handle,
+        const TPM2B_DIGEST *digest, const TPMT_SIGNATURE *signature,
+        TPMT_TK_VERIFIED **validation) {
+
+    TSS2_RC rval = Esys_VerifySignature(ectx,
+            key_handle, ESYS_TR_NONE, ESYS_TR_NONE,
+            ESYS_TR_NONE, digest, signature, validation);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_VerifySignature, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_readclock(ESYS_CONTEXT *ectx, TPMS_TIME_INFO **current_time) {
+
+    TSS2_RC rval = Esys_ReadClock(ectx,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
+            current_time);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_ReadClock, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_setclock(ESYS_CONTEXT *ectx, tpm2_loaded_object *object, UINT64 new_time) {
+
+    ESYS_TR shandle1 = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx,
+            object->tr_handle, object->session, &shandle1);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Couldn't get shandle for lockout hierarchy");
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_ClockSet(ectx,
+            object->tr_handle,
+            shandle1, ESYS_TR_NONE, ESYS_TR_NONE,
+            new_time);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_ClockSet, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_clockrateadjust(ESYS_CONTEXT *ectx, tpm2_loaded_object *object,
+        TPM2_CLOCK_ADJUST rate_adjust) {
+
+    ESYS_TR shandle1 = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx,
+            object->tr_handle, object->session, &shandle1);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Couldn't get shandle for lockout hierarchy");
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_ClockRateAdjust(ectx,
+            object->tr_handle,
+            shandle1, ESYS_TR_NONE, ESYS_TR_NONE,
+            rate_adjust);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_ClockRateAdjust, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_shutdown(ESYS_CONTEXT *ectx, TPM2_SU shutdown_type) {
+
+    TSS2_RC rval = Esys_Shutdown(ectx,
+            ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
+            shutdown_type);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_Shutdown, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_gettime(ESYS_CONTEXT *ectx,
+        tpm2_loaded_object *privacy_admin,
+        tpm2_loaded_object *signing_object,
+        const TPM2B_DATA *qualifying_data,
+        const TPMT_SIG_SCHEME *scheme,
+        TPM2B_ATTEST **time_info,
+        TPMT_SIGNATURE **signature) {
+
+    ESYS_TR privacy_admin_session_handle = ESYS_TR_NONE;
+    tool_rc rc = tpm2_auth_util_get_shandle(ectx,
+            privacy_admin->tr_handle, privacy_admin->session, &privacy_admin_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Couldn't get shandle for privacy admin");
+        return rc;
+    }
+
+    ESYS_TR sign_session_handle = ESYS_TR_NONE;
+    rc = tpm2_auth_util_get_shandle(ectx,
+            signing_object->tr_handle, signing_object->session, &sign_session_handle);
+    if (rc != tool_rc_success) {
+        LOG_ERR("Couldn't get shandle for signing key");
+        return rc;
+    }
+
+    TSS2_RC rval = Esys_GetTime(
+            ectx,
+            privacy_admin->tr_handle,
+            signing_object->tr_handle,
+            privacy_admin_session_handle,
+            sign_session_handle,
+            ESYS_TR_NONE,
+            qualifying_data,
+            scheme,
+            time_info,
+            signature);
+    if (rval != TPM2_RC_SUCCESS) {
+        LOG_PERR(Esys_GetTime, rval);
+        return tool_rc_from_tpm(rval);
+    }
+
+    return tool_rc_success;
+}
