@@ -11,6 +11,7 @@
 #include "tpm2_openssl.h"
 #include "tpm2_policy.h"
 #include "tpm2_tool.h"
+#include "tpm2_util.h"
 
 static bool evaluate_populate_pcr_digests(TPML_PCR_SELECTION *pcr_selections,
         const char *raw_pcrs_file, TPML_DIGEST *pcr_values) {
@@ -156,19 +157,16 @@ tool_rc tpm2_policy_build_pcr(ESYS_CONTEXT *ectx, tpm2_session *policy_session,
 
 tool_rc tpm2_policy_build_policyauthorize(ESYS_CONTEXT *ectx,
         tpm2_session *policy_session, const char *policy_digest_path,
-        const char *policy_qualifier_path,
+        const char *qualifying_data,
         const char *verifying_pubkey_name_path, const char *ticket_path) {
 
-    unsigned long file_size = 0;
-
-    bool result = files_get_file_size_path(policy_digest_path, &file_size);
-    if (!result) {
-        return tool_rc_general_error;
-    }
-
-    TPM2B_DIGEST approved_policy = { .size = (uint16_t) file_size };
-    result = files_load_bytes_from_path(policy_digest_path,
+    bool result = true;
+    TPM2B_DIGEST approved_policy = { .size = 0 };
+    if (policy_digest_path) {
+        approved_policy.size = UINT16_MAX;
+        result = files_load_bytes_from_path(policy_digest_path,
             approved_policy.buffer, &approved_policy.size);
+    }
     if (!result) {
         return tool_rc_general_error;
     }
@@ -176,24 +174,19 @@ tool_rc tpm2_policy_build_policyauthorize(ESYS_CONTEXT *ectx,
     /*
      * Qualifier data is optional. If not specified default to 0
      */
-    file_size = 0;
-    if (policy_qualifier_path) {
-        result = files_get_file_size_path(policy_qualifier_path, &file_size);
-        if (!result) {
-            return tool_rc_general_error;
-        }
-    }
 
-    TPM2B_NONCE policy_qualifier = { .size = (uint16_t) file_size };
+    TPM2B_NONCE policy_qualifier = { .size = 0 };
 
-    if (file_size != 0) {
-        result = files_load_bytes_from_path(policy_qualifier_path,
-                policy_qualifier.buffer, &policy_qualifier.size);
+    if (qualifying_data) {
+        policy_qualifier.size = sizeof(policy_qualifier.buffer);
+        result = tpm2_util_bin_from_hex_or_file(qualifying_data,
+                &policy_qualifier.size, policy_qualifier.buffer);
         if (!result) {
             return tool_rc_general_error;
         }
     }
 
+    unsigned long file_size = 0;
     result = files_get_file_size_path(verifying_pubkey_name_path, &file_size);
     if (!result) {
         return tool_rc_general_error;
@@ -247,12 +240,209 @@ tool_rc tpm2_policy_build_policypassword(ESYS_CONTEXT *ectx,
             ESYS_TR_NONE, ESYS_TR_NONE);
 }
 
+tool_rc tpm2_policy_build_policynamehash(ESYS_CONTEXT *ectx,
+    tpm2_session *session, const TPM2B_DIGEST *name_hash) {
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(session);
+
+    return tpm2_policy_namehash(ectx, policy_session_handle, name_hash);
+}
+
+tool_rc tpm2_policy_build_policytemplate(ESYS_CONTEXT *ectx,
+    tpm2_session *session, const TPM2B_DIGEST *template_hash) {
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(session);
+
+    return tpm2_policy_template(ectx, policy_session_handle, template_hash);
+}
+
+tool_rc tpm2_policy_build_policyauthvalue(ESYS_CONTEXT *ectx,
+        tpm2_session *session) {
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(session);
+
+    return tpm2_policy_authvalue(ectx, policy_session_handle, ESYS_TR_NONE,
+            ESYS_TR_NONE, ESYS_TR_NONE);
+}
+
 tool_rc tpm2_policy_build_policysecret(ESYS_CONTEXT *ectx,
-        tpm2_session *policy_session, tpm2_loaded_object *auth_entity_obj) {
+        tpm2_session *policy_session, tpm2_loaded_object *auth_entity_obj,
+        INT32 expiration, TPMT_TK_AUTH **policy_ticket,
+        TPM2B_TIMEOUT **timeout, bool is_nonce_tpm,
+        const char *policy_qualifier_data) {
+
+    /*
+     * Qualifier data is optional. If not specified default to 0
+     */
+    TPM2B_NONCE policy_qualifier = TPM2B_EMPTY_INIT;
+    if (policy_qualifier_data) {
+        policy_qualifier.size = sizeof(policy_qualifier.buffer);
+        bool result = tpm2_util_bin_from_hex_or_file(policy_qualifier_data,
+                &policy_qualifier.size,
+                policy_qualifier.buffer);
+        if (!result) {
+            return tool_rc_general_error;
+        }
+    }
 
     ESYS_TR policy_session_handle = tpm2_session_get_handle(policy_session);
 
-    return tpm2_policy_secret(ectx, auth_entity_obj, policy_session_handle);
+    TPM2B_NONCE *nonce_tpm = NULL;
+    tool_rc rc = tool_rc_success;
+    if (is_nonce_tpm) {
+        rc = tpm2_sess_get_noncetpm(ectx, policy_session_handle, &nonce_tpm);
+        if (rc != tool_rc_success) {
+            goto tpm2_policy_build_policysecret_out;
+        }
+    }
+
+    rc = tpm2_policy_secret(ectx, auth_entity_obj, policy_session_handle,
+        expiration, policy_ticket, timeout, nonce_tpm, &policy_qualifier);
+
+tpm2_policy_build_policysecret_out:
+    Esys_Free(nonce_tpm);
+    return rc;
+}
+
+tool_rc tpm2_policy_build_policyticket(ESYS_CONTEXT *ectx,
+    tpm2_session *policy_session, char *policy_timeout_path,
+    const char *qualifier_data, char *policy_ticket_path,
+    const char *auth_name_path) {
+
+    unsigned long file_size = 0;
+
+    bool result = files_get_file_size_path(policy_timeout_path, &file_size);
+    if (!result) {
+        return tool_rc_general_error;
+    }
+    TPM2B_TIMEOUT policy_timeout = { .size = (uint16_t) file_size };
+    if (policy_timeout.size) {
+        result = files_load_bytes_from_path(policy_timeout_path,
+                policy_timeout.buffer, &policy_timeout.size);
+        if (!result) {
+            return tool_rc_general_error;
+        }
+    }
+
+    result = files_get_file_size_path(auth_name_path, &file_size);
+    if (!result) {
+        return tool_rc_general_error;
+    }
+    TPM2B_NAME auth_name = { .size = (uint16_t) file_size };
+    if (auth_name.size) {
+        result = files_load_bytes_from_path(auth_name_path,
+                auth_name.name, &auth_name.size);
+        if (!result) {
+            return tool_rc_general_error;
+        }
+    }
+
+    TPM2B_NONCE policyref = TPM2B_EMPTY_INIT;
+    if (qualifier_data) {
+        policyref.size = sizeof(policyref.buffer);
+        result = tpm2_util_bin_from_hex_or_file(qualifier_data, &policyref.size,
+                policyref.buffer);
+        if (!result) {
+            return tool_rc_general_error;
+        }
+    }
+
+    TPMT_TK_AUTH ticket = { 0 };
+    result = files_load_authorization_ticket(policy_ticket_path, &ticket);
+    if (!result) {
+        LOG_ERR("Failed loading authorization ticket.");
+        return tool_rc_general_error;
+    }
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(policy_session);
+
+    return tpm2_policy_ticket(ectx, policy_session_handle, &policy_timeout,
+        &policyref, &auth_name, &ticket);
+
+    return tool_rc_success;
+}
+
+tool_rc tpm2_policy_build_policysigned(ESYS_CONTEXT *ectx,
+        tpm2_session *policy_session, tpm2_loaded_object *auth_entity_obj,
+        TPMT_SIGNATURE *signature, INT32 expiration, TPM2B_TIMEOUT **timeout,
+        TPMT_TK_AUTH **policy_ticket, const char *policy_qualifier_data,
+        bool is_nonce_tpm, const char *raw_data_path) {
+
+    bool result = true;
+
+    /*
+     * Qualifier data is optional. If not specified default to 0
+     */
+    TPM2B_NONCE policy_qualifier = TPM2B_EMPTY_INIT;
+
+    if (policy_qualifier_data) {
+        policy_qualifier.size = sizeof(policy_qualifier.buffer);
+        result = tpm2_util_bin_from_hex_or_file(policy_qualifier_data,
+                &policy_qualifier.size,
+                policy_qualifier.buffer);
+        if (!result) {
+            return tool_rc_general_error;
+        }
+    }
+
+    ESYS_TR policy_session_handle = tpm2_session_get_handle(policy_session);
+
+    TPM2B_NONCE *nonce_tpm = NULL;
+    tool_rc rc = tool_rc_success;
+    if (is_nonce_tpm) {
+        rc = tpm2_sess_get_noncetpm(ectx, policy_session_handle, &nonce_tpm);
+        if (rc != tool_rc_success) {
+            goto tpm2_policy_build_policysigned_out;
+        }
+    }
+
+    /*
+     * TPM-Rev-2.0-Part-3-Commands-01.38.pdf
+     * aHash ≔ HauthAlg(nonceTPM || expiration || cpHashA || policyRef)
+     */
+    if (raw_data_path) {
+        uint16_t raw_data_len = (nonce_tpm ? nonce_tpm->size : 0) +
+            sizeof(INT32) + policy_qualifier.size;
+
+        uint8_t *raw_data = malloc(raw_data_len);
+        if (!raw_data) {
+            LOG_ERR("oom");
+            rc = tool_rc_general_error;
+            goto tpm2_policy_build_policysigned_out;
+        }
+        //nonceTPM
+        uint16_t offset = 0;
+        if (nonce_tpm) {
+            memcpy(raw_data, nonce_tpm->buffer, nonce_tpm->size);
+            offset += nonce_tpm->size;
+        }
+        //expiration
+        UINT32 endswap_data = tpm2_util_endian_swap_32(expiration);
+        memcpy(raw_data + offset, (UINT8 *)&endswap_data, sizeof(INT32));
+        offset += sizeof(INT32);
+        //policyRef
+        memcpy(raw_data + offset, policy_qualifier.buffer,
+            policy_qualifier.size);
+
+        bool result = files_save_bytes_to_file(raw_data_path, raw_data,
+            raw_data_len);
+        free(raw_data);
+        if (!result) {
+            rc = tool_rc_general_error;
+        }
+        /*
+         * We return since we only need to generate the raw signing data
+         */
+        goto tpm2_policy_build_policysigned_out;
+    }
+
+    rc = tpm2_policy_signed(ectx, auth_entity_obj, policy_session_handle,
+        signature, expiration, timeout, policy_ticket, &policy_qualifier,
+        nonce_tpm);
+
+tpm2_policy_build_policysigned_out:
+    Esys_Free(nonce_tpm);
+    return rc;
 }
 
 tool_rc tpm2_policy_get_digest(ESYS_CONTEXT *ectx, tpm2_session *session,
@@ -273,6 +463,15 @@ tool_rc tpm2_policy_build_policycommandcode(ESYS_CONTEXT *ectx,
             ESYS_TR_NONE, command_code);
 }
 
+tool_rc tpm2_policy_build_policynvwritten(ESYS_CONTEXT *ectx,
+        tpm2_session *session, TPMI_YES_NO written_set) {
+
+    ESYS_TR handle = tpm2_session_get_handle(session);
+
+    return tpm2_policy_nv_written(ectx, handle, ESYS_TR_NONE, ESYS_TR_NONE,
+            ESYS_TR_NONE, written_set);
+}
+
 tool_rc tpm2_policy_build_policylocality(ESYS_CONTEXT *ectx,
         tpm2_session *session, TPMA_LOCALITY locality) {
 
